@@ -1,10 +1,10 @@
 /*  Part of SWI-Prolog
-
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2004-2017, University of Amsterdam
+    Copyright (c)  2004-2018, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -86,10 +86,6 @@ Address Converstion
 	nbio_get_sockaddr()
 	nbio_get_ip4()
 
-Waiting
-
-	nbio_select()
-
 Alternative to nbio_read() and nbio_write(), the application program may
 call  the  low-level  I/O  routines  in    non-blocking  mode  and  call
 nbio_wait(int socket, nbio_request request). This  function returns 0 if
@@ -102,11 +98,10 @@ Windows issues
 --------------
 
 Winsock is hard to handle in blocking   mode  without blocking the whole
-lot, notably (timeout) signals. We  therefore   have  a  seperate thread
-dealing with I/O and  operating   the  sockets  through WSAAsyncSelect()
-generated events. Requests are registered   with the plsocket structure,
-handled and handled in the socket thread.  Upon completion, a message is
-sent back to the waiting thread.
+lot, notably (timeout) signals. Old versions uses WSAASyncSelect() and a
+separate thread. This is slow and   complicated. The current version was
+written     by     Keri     Harris     and     uses     WSAEvent     and
+MsgWaitForMultipleObjects(). See wait_socket().
 
 Unix issues
 -----------
@@ -115,9 +110,7 @@ In the Unix version we simply call PL_dispatch() before doing recv() and
 leave the details to this function.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#ifdef HAVE_CONFIG_H
 #include <config.h>
-#endif
 
 #if defined(__MINGW32__)
 #define __try
@@ -174,63 +167,19 @@ leave the details to this function.
 #define SOCKET_ERROR (-1)
 #endif
 
-#ifdef _REENTRANT
-#if __WINDOWS__
-static CRITICAL_SECTION mutex;
-
-#define LOCK()			EnterCriticalSection(&mutex)
-#define UNLOCK()		LeaveCriticalSection(&mutex)
-#define LOCK_FREE()		(void)0
-#define UNLOCK_FREE()		(void)0
-#define INITLOCK()		InitializeCriticalSection(&mutex)
-#define LOCK_SOCKET(s)		(void)0
-#define UNLOCK_SOCKET(s)	(void)0
-#define INIT_SOCKET_LOCK(s)	(void)0
-#define FREE_SOCKET_LOCK(s)	(void)0
-#else /*__WINDOWS__*/
-#include <pthread.h>
-
-static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
-
-#define LOCK()			pthread_mutex_lock(&mutex)
-#define UNLOCK()		pthread_mutex_unlock(&mutex)
-#define LOCK_FREE()		(void)0
-#define UNLOCK_FREE()		(void)0
-#define INITLOCK()		(void)0
-#define FREE_SOCKET_LOCK(s)	(void)0
-#endif /*__WINDOWS__*/
-#else /* _REENTRANT */
-#define LOCK()			(void)0
-#define UNLOCK()		(void)0
-#define LOCK_FREE()		(void)0
-#define UNLOCK_FREE()		(void)0
-#define INITLOCK()		(void)0
-#define FREE_SOCKET_LOCK(s)	(void)0
-#if __WINDOWS__
-#define LOCK_SOCKET(s)		(void)0
-#define UNLOCK_SOCKET(s)	(void)0
-#define INIT_SOCKET_LOCK(s)	(void)0
-#endif
-#endif /* _REENTRANT */
-
 #define set(s, f)   ((s)->flags |= (f))
 #define clear(s, f) ((s)->flags &= ~(f))
 #define true(s, f)  ((s)->flags & (f))
 #define false(s, f) (!true(s, f))
 
-#define PLSOCK_MAGIC 0x38da3f2c
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-NOTE: We must lock  the  structure   to  avoid  freeSocket() called from
-Prolog deleting the socket while there are   still  pending events on it
-that are concurrently executed in the socket thread.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+#define PLSOCK_MAGIC  0x38da3f2c
+#define PLSOCK_CMAGIC 0x38da3f2d
 
 typedef struct _plsocket
 { int		    magic;		/* PLSOCK_MAGIC */
-  nbio_sock_t	    id;			/* Integer id */
   SOCKET	    socket;		/* The OS socket */
   int		    flags;		/* Misc flags */
+  atom_t	    symbol;		/* <socket>(%p) */
   IOSTREAM *	    input;		/* input stream */
   IOSTREAM *	    output;		/* output stream */
 #ifdef __WINDOWS__
@@ -238,17 +187,26 @@ typedef struct _plsocket
 #endif
 } plsocket;
 
+#define VALID_SOCKET_RET(s, r) \
+	do						\
+	{ if ( !(s && (s)->magic == PLSOCK_MAGIC) )	\
+	  { errno = EINVAL;				\
+	    return (r);					\
+	  }						\
+	} while(0)
+#define VALID_SOCKET(s) VALID_SOCKET_RET(s, -1)
+
+
 static plsocket *allocSocket(SOCKET socket);
 #ifdef __WINDOWS__
 static const char *WinSockError(unsigned long eno);
 #endif
 static int need_retry(int error);
-static int freeSocket(plsocket *s);
 
 #ifdef O_DEBUG
 static int debugging;
 
-NBIO_EXPORT(int)
+int
 nbio_debug(int level)
 { int old = debugging;
 
@@ -267,7 +225,7 @@ nbio_debug(int level)
 { return 0;
 }
 
-#endif
+#endif /*O_DEBUG*/
 
 		 /*******************************
 		 *	  COMPATIBILITY		*
@@ -306,17 +264,14 @@ event_name(int ev)
   return strdup(buf);
 }
 
-#endif
+#endif /*O_DEBUG*/
 
 #define F_SETFL		0
 #define O_NONBLOCK	0
 
 static int
 nbio_fcntl(nbio_sock_t socket, int op, int arg)
-{ plsocket *s;
-
-  if ( !(s=nbio_to_plsocket(socket)) )
-    return -1;
+{ VALID_SOCKET(socket);
 
   switch(op)
   { case F_SETFL:
@@ -330,9 +285,9 @@ nbio_fcntl(nbio_sock_t socket, int op, int arg)
 	  int non_block;
 #endif
 	  non_block = 1;
-	  rval = ioctlsocket(s->socket, FIONBIO, &non_block);
+	  rval = ioctlsocket(socket->socket, FIONBIO, &non_block);
 	  if ( rval )
-	  { s->flags |= PLSOCK_NONBLOCK;
+	  { set(socket, PLSOCK_NONBLOCK);
 	    return 0;
 	  }
 
@@ -359,6 +314,14 @@ need_retry(int error)
 }
 
 
+/*  wait_socket() waits for the socket to become ready and in the
+ *  while waiting processes Windows messages, including Prolog
+ *  timeout.
+ *
+ *  Returns TRUE if all ok and FALSE if some Prolog exception has
+ *  been raised.
+ */
+
 static int
 wait_socket(plsocket *s)
 { int index;
@@ -416,7 +379,9 @@ wait_socket(plsocket *s)
         DispatchMessage(&msg);
 
         if ( PL_handle_signals() < 0 )
+	{ errno = EPLEXCEPTION;
           return FALSE;
+	}
         continue;
       }
     }
@@ -427,156 +392,12 @@ wait_socket(plsocket *s)
 
 int
 nbio_wait(nbio_sock_t socket, nbio_request request)
-{ plsocket *s;
+{ VALID_SOCKET(socket);
 
-  if ( !(s=nbio_to_plsocket(socket)) )
-    return -1;
-
-  return wait_socket(s) ? 0 : -1;
+  return wait_socket(socket) ? 0 : -1;
 }
 
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-nbio_select() selects using a set of socket streams.
-
-NOTE: The Windows versions uses our   nbio_sock_t abstraction, while the
-other version uses  the  raw  Unix   file  descriptors  referencing  the
-underlying socket.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-int
-nbio_select(int n,
-	    fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
-	    struct timeval *timeout)
-{ SOCKET sockets[WSA_MAXIMUM_WAIT_EVENTS];
-  WSAEVENT events[WSA_MAXIMUM_WAIT_EVENTS];
-  int fds[WSA_MAXIMUM_WAIT_EVENTS];
-  int i, index;
-  int count = 0;
-  int n_ready = 0;
-  DWORD t_end;
-
-  if ( readfds )
-  { for(i=0; i<n; i++)
-    { if ( FD_ISSET(i, readfds) )
-      { plsocket *s = nbio_to_plsocket(i);
-
-	if ( s )
-        { WSANETWORKEVENTS network_events;
-          WSAEVENT event = WSACreateEvent();
-          int event_type = (s->flags & PLSOCK_LISTEN) ? FD_ACCEPT : FD_READ;
-
-          WSAEventSelect(s->socket, event, event_type);
-          fds[count] = i;
-          sockets[count] = s->socket;
-          events[count] = event;
-          count++;
-
-          if ( WSAEnumNetworkEvents(s->socket, NULL, &network_events) == SOCKET_ERROR )
-          { nbio_error(GET_ERRNO, TCP_ERRNO);
-            return -1;
-          }
-          DEBUG(2,
-                if ( network_events.lNetworkEvents )
-                { char *nm = event_name(network_events.lNetworkEvents);
-                  Sdprintf("WM_SOCKET on %d: ev=(%s)\n", s->socket, nm);
-	          free(nm);
-                });
-          if ( (network_events.lNetworkEvents & FD_ACCEPT) &&
-               !network_events.iErrorCode[FD_ACCEPT_BIT] )
-          { n_ready++;
-          } else if ( (network_events.lNetworkEvents & FD_READ) &&
-		      !network_events.iErrorCode[FD_READ_BIT] )
-          { n_ready++;
-          } else
-          { FD_CLR(i, readfds);
-          }
-        } else
-        { DEBUG(2, Sdprintf("nbio_select(): no socket for %d\n", i));
-          FD_CLR(i, readfds);
-	}
-      }
-    }
-  }
-
-  if ( writefds )
-    return -1;				/* not yet implemented */
-  if ( exceptfds )
-    return -1;				/* idem (might never be) */
-
-  if ( timeout )
-  { t_end = GetTickCount();
-    t_end += timeout->tv_sec*1000;
-    t_end += timeout->tv_usec/1000;
-  }
-
-  if ( n_ready )
-  { return n_ready;
-  }
-
-  FD_ZERO(readfds);
-
-  for(;;)
-  { DWORD t, msecs;
-
-    DEBUG(2, Sdprintf("waiting on %d sockets\n", count));
-
-    if ( timeout )
-    { t = GetTickCount();
-      msecs = t_end - t;
-      if ( msecs < 0 )
-        msecs = -msecs; /* wrapped around */
-    } else
-    { msecs = INFINITE;
-    }
-
-    index = MsgWaitForMultipleObjects(count, events, FALSE, msecs, QS_ALLINPUT);
-
-    if ( index == WAIT_TIMEOUT )
-    { DEBUG(2, Sdprintf("nbio_select() timed out\n"));
-      return 0;
-    }
-
-    if ( index == WAIT_FAILED )
-    { nbio_error(GET_ERRNO, TCP_ERRNO);
-      return -1;
-    } else if ( index < WAIT_OBJECT_0+count ) /* socket event */
-    { WSANETWORKEVENTS network_events;
-      if ( WSAEnumNetworkEvents(sockets[index-WSA_WAIT_EVENT_0], NULL, &network_events) == SOCKET_ERROR )
-      { nbio_error(GET_ERRNO, TCP_ERRNO);
-        return -1;
-      }
-      DEBUG(2,
-            { char *nm = event_name(network_events.lNetworkEvents);
-              Sdprintf("WM_SOCKET on %d: ev=(%s)\n", sockets[index-WSA_WAIT_EVENT_0], nm);
-	      free(nm);
-            });
-      if ( (network_events.lNetworkEvents & FD_ACCEPT) &&
-           !network_events.iErrorCode[FD_ACCEPT_BIT] )
-      { FD_SET(fds[index-WSA_WAIT_EVENT_0], readfds);
-        return 1;
-      } else if ( (network_events.lNetworkEvents & FD_READ) &&
-		  !network_events.iErrorCode[FD_READ_BIT] )
-      { FD_SET(fds[index-WSA_WAIT_EVENT_0], readfds);
-        return 1;
-      }
-    } else if ( index == WAIT_OBJECT_0+count ) /* message event */
-    { MSG msg;
-      DEBUG(2, Sdprintf("nbio_select() interrupted\n"));
-      while( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) )
-      { TranslateMessage(&msg);
-        DispatchMessage(&msg);
-
-        if ( PL_handle_signals() < 0 )
-          return -1;
-        continue;
-      }
-    }
-  }
-
-  return -1;
-}
-
 #else /*__WINDOWS__*/
 
 static int
@@ -628,7 +449,11 @@ wait_socket(plsocket *s)
       }
 #endif
     } else
-    { return PL_dispatch(fd, PL_DISPATCH_WAIT);
+    { int rc;
+
+      if ( !(rc = PL_dispatch(fd, PL_DISPATCH_WAIT)) )
+	errno = EPLEXCEPTION;
+      return rc;
     }
   }
 
@@ -638,28 +463,23 @@ wait_socket(plsocket *s)
 
 int
 nbio_wait(nbio_sock_t socket, nbio_request request)
-{ plsocket *s;
+{ VALID_SOCKET(socket);
 
-  if ( !(s=nbio_to_plsocket(socket)) )
-    return -1;
-
-  return wait_socket(s) ? 0 : -1;
+  return wait_socket(socket) ? 0 : -1;
 }
 
 
 static int
 nbio_fcntl(nbio_sock_t socket, int op, int arg)
-{ plsocket *s;
-  int rc;
+{ int rc;
 
-  if ( !(s=nbio_to_plsocket(socket)) )
-    return -1;
+  VALID_SOCKET(socket);
 
-  rc = fcntl(s->socket, op, arg);
+  rc = fcntl(socket->socket, op, arg);
 
   if ( rc == 0 )
   { if ( op == F_SETFL && arg == O_NONBLOCK )
-      s->flags |= PLSOCK_NONBLOCK;
+      set(socket, PLSOCK_NONBLOCK);
   } else
     nbio_error(GET_ERRNO, TCP_ERRNO);
 
@@ -680,73 +500,24 @@ static atom_t ATOM_any;
 static atom_t ATOM_broadcast;
 static atom_t ATOM_loopback;
 
-static plsocket **sockets = NULL;	/* id --> plsocket* */
-static size_t	socks_count = 0;	/* #registered sockets */
-static size_t	socks_allocated = 0;	/* #allocated entries */
 static int initialised = FALSE;		/* Windows only */
 
-
-static plsocket *
-nbio_to_plsocket_nolock(nbio_sock_t socket)
-{ plsocket *p;
-
-  if ( socket < 0 || (size_t)socket >= socks_allocated )
-  { errno = EINVAL;
-    return NULL;
-  }
-
-  p = sockets[socket];
-
-  if ( !p || p->magic != PLSOCK_MAGIC )
-  { DEBUG(1, Sdprintf("Invalid NBIO socket: %d\n", socket));
-    errno = EINVAL;
-    return NULL;
-  }
-
-  return p;
-}
-
-
 SOCKET
-plsocket_handle(plsocket_ptr pls)
-{ return pls->socket;
-}
-
-
-static plsocket *
-nbio_to_plsocket_raw(nbio_sock_t socket)
-{ plsocket *s;
-
-  LOCK();
-  s = nbio_to_plsocket_nolock(socket);
-  UNLOCK();
-
-  return s;
-}
-
-
-plsocket *
-nbio_to_plsocket(nbio_sock_t socket)
-{ plsocket *p;
-
-  if ( !(p=nbio_to_plsocket_raw(socket)) )
-    return NULL;
-
-  return p;
-}
-
-
-NBIO_EXPORT(SOCKET)
 nbio_fd(nbio_sock_t socket)
-{ plsocket *p;
+{ VALID_SOCKET(socket);
 
-  if ( !(p=nbio_to_plsocket_nolock(socket)) )
-    return -1;
-
-  return p->socket;
+  return socket->socket;
 }
 
+void
+nbio_set_symbol(nbio_sock_t socket, atom_t symbol)
+{ socket->symbol = symbol;
+}
 
+int
+is_nbio_socket(nbio_sock_t socket)
+{ return socket && socket->magic == PLSOCK_MAGIC;
+}
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Allocate a wrapper for an OS  socket.   The  wrapper  is allocated in an
@@ -756,38 +527,14 @@ FD_SET, etc. for implementing a compatible nbio_select().
 
 static plsocket *
 allocSocket(SOCKET socket)
-{ plsocket *p = NULL;
-  size_t i;
+{ plsocket *p;
 
-  LOCK();
-  if ( socks_count+1 > socks_allocated )
-  { if ( socks_allocated )
-    { size_t newa = socks_allocated*2;
-
-      sockets = PL_realloc(sockets, sizeof(plsocket*)*newa);
-      for(i=socks_allocated; i<newa; i++)
-	sockets[i] = NULL;
-      socks_allocated = newa;
-    } else
-    { socks_allocated = 32;
-      sockets = PL_malloc(sizeof(plsocket*)*socks_allocated);
-      memset(sockets, 0, sizeof(plsocket*)*socks_allocated);
-    }
+  if ( !(p = malloc(sizeof(*p))) )
+  { PL_resource_error("memory");
+    return NULL;
   }
 
-  for(i=0; i<socks_allocated; i++)
-  { if ( (p=sockets[i]) == NULL )
-    { sockets[i] = p = PL_malloc(sizeof(*p));
-      socks_count++;
-      break;
-    }
-  }
-  UNLOCK();
-
-  assert(i<socks_allocated);
-
   memset(p, 0, sizeof(*p));
-  p->id     = (int)i;			/* place in the array */
   p->socket = socket;
   p->flags  = PLSOCK_DISPATCH|PLSOCK_VIRGIN;	/* by default, dispatch */
   p->magic  = PLSOCK_MAGIC;
@@ -800,38 +547,44 @@ allocSocket(SOCKET socket)
   }
 #endif
 
-  DEBUG(2, Sdprintf("[%d]: allocSocket(%d): bound to %d\n",
-		    PL_thread_self(), socket, p->id));
+  DEBUG(2, Sdprintf("[%d]: allocSocket(%d) --> %p\n",
+		    PL_thread_self(), socket, p));
 
   return p;
 }
 
 
+void
+freeSocket(nbio_sock_t s)
+{ if ( s )
+  { if ( s->magic == PLSOCK_CMAGIC )
+      free(s);
+    else
+      s->symbol = 0;
+  }
+}
+
+
 static int
-freeSocket(plsocket *s)
+closeSocket(plsocket *s)
 { int rval;
-  nbio_sock_t socket;
   SOCKET sock;
 
-  DEBUG(2, Sdprintf("Closing %p (%d)\n", s, s ? s->id : 0));
+  DEBUG(2, Sdprintf("Closing %p (%d)\n", s, s->socket));
   if ( !s || s->magic != PLSOCK_MAGIC )
-  { DEBUG(1, Sdprintf("OOPS: freeSocket(%p) s->magic = %ld\n", s, s ? s->magic : 0));
+  { DEBUG(1, Sdprintf("OOPS: closeSocket(%p) s->magic = %ld\n",
+		      s, s ? s->magic : 0));
     errno = EINVAL;
     return -1;
   }
 
-  LOCK_FREE();
-  LOCK();
-  sockets[s->id] = NULL;
-  socks_count--;
-  UNLOCK();
-
   sock = s->socket;
-  socket = s->id;
-  s->magic = 0;
-  FREE_SOCKET_LOCK(s);
-  PL_free(s);
-  UNLOCK_FREE();
+  s->magic = PLSOCK_CMAGIC;
+
+#ifdef __WINDOWS__
+  if ( s->event )
+    WSACloseEvent(s->event);
+#endif
 
   if ( sock != INVALID_SOCKET )
   { again:
@@ -839,14 +592,17 @@ freeSocket(plsocket *s)
     { if ( errno == EINTR )
 	goto again;
     }
-    DEBUG(2, Sdprintf("freeSocket(%d=%d): closesocket() returned %d\n",
-		      socket, (int)sock, rval));
+    DEBUG(2, Sdprintf("closeSocket(%p=%d): closesocket() returned %d\n",
+		      s, (int)sock, rval));
   } else
-  { DEBUG(2, Sdprintf("freeSocket(%d=%d): already closed\n",
-		      socket, (int)sock));
+  { DEBUG(2, Sdprintf("closeSocket(%p=%d): already closed\n",
+		      s, (int)sock));
     rval = 0;				/* already closed.  Use s->error? */
   }
 
+  if ( !s->symbol )
+    free(s);
+
   return rval;
 }
 
@@ -960,95 +716,81 @@ WinSockError(unsigned long error)
 #endif /*BILLY_GETS_BETTER*/
 #endif /*__WINDOWS__*/
 
-#ifdef HAVE_H_ERRNO
-typedef struct
-{ int code;
-  const char *string;
-} error_codes;
+		 /*******************************
+		 *     POSIX SOCKET ERRORS	*
+		 *******************************/
 
-static error_codes h_errno_codes[] = {
-#ifdef HOST_NOT_FOUND
-    { HOST_NOT_FOUND, "Host not found" },
-#endif
-#ifdef TRY_AGAIN
-    { TRY_AGAIN, "Try Again" },
-#endif
-#ifdef NO_RECOVERY
-    { NO_RECOVERY, "No Recovery" },
-#endif
-#ifdef NO_DATA
-    { NO_DATA, "No Data" },
-#endif
-#ifdef NO_ADDRESS
-    { NO_ADDRESS, "No Address" },
-#endif
-    {0, NULL}
-};
+#include "esymbols.ic"
 
-#else /*HAVE_H_ERRNO*/
-#define h_errno_codes NULL
-typedef void * error_codes;
-#endif /*HAVE_H_ERRNO*/
+#ifndef __WINDOWS__
+#ifndef HAVE_HSTRERROR
+#define hstrerror my_hstrerror
+
+static const char *
+hstrerror(int code)
+{ return error_symbol(code, h_errno_symbols);
+}
+#endif /*HAVE_HSTRERROR*/
+
+#ifndef HAVE_GAI_STRERROR
+#define gai_strerror my_gai_strerror
+
+static const char *
+gai_strerror(int code)
+{ return error_symbol(code, gai_errno_symbols);
+}
+#endif /*HAVE_HAVE_GAI_STRERROR*/
+#endif /*__WINDOWS__*/
 
 int
 nbio_error(int code, nbio_error_map mapid)
 { const char *msg;
-  term_t except = PL_new_term_ref();
-#ifdef HAVE_H_ERRNO
-  error_codes *map;
-#endif
+  const char *symbol;
+  term_t ex;
 
-  if ( code == EPLEXCEPTION )
+  if ( code == EPLEXCEPTION || PL_exception(0) )
     return FALSE;
 
-#ifdef HAVE_H_ERRNO
-  switch( mapid )
-  { case TCP_HERRNO:
-      map = h_errno_codes;
-      break;
-    default:
-      map = NULL;
-  }
-#endif
-
-  {
 #ifdef __WINDOWS__
   msg = WinSockError(code);
+  symbol = error_symbol(code, wsa_errno_symbols);
 #else
-
-#ifdef HAVE_H_ERRNO
-  static char msgbuf[100];
-
-  if ( map )
-  { while( map->code && map->code != code )
-      map++;
-    if ( map->code )
-      msg = map->string;
-    else
-    { sprintf(msgbuf, "Unknown error %d", code);
-      msg = msgbuf;
-    }
-  } else
-#endif
-    msg = strerror(code);
-#endif /*__WINDOWS__*/
-
-  if ( !PL_unify_term(except,
-		      CompoundArg("error", 2),
-		        CompoundArg("socket_error", 1),
-		          AtomArg(msg),
-		        PL_VARIABLE) )
-    return FALSE;
+  switch( mapid )
+  { case TCP_ERRNO:
+      msg = strerror(code);
+      symbol = error_symbol(code, errno_symbols);
+      break;
+    case TCP_HERRNO:
+      msg = hstrerror(code);
+      symbol = error_symbol(code, h_errno_symbols);
+      break;
+    case TCP_GAI_ERRNO:
+      msg = gai_strerror(code);
+      symbol = error_symbol(code, gai_errno_symbols);
+      break;
+    default:
+      assert(0);
+      msg = NULL;
+      break;
   }
+#endif
 
-  return PL_raise_exception(except);
+  errno = EPLEXCEPTION;
+  return ( (ex = PL_new_term_ref()) &&
+	   PL_unify_term(ex,
+			 CompoundArg("error", 2),
+			   CompoundArg("socket_error", 2),
+			     AtomArg(symbol),
+			     AtomArg(msg),
+			   PL_VARIABLE) &&
+	   PL_raise_exception(ex)
+	 );
 }
 
 
 const char *
 nbio_last_error(nbio_sock_t socket)
-{
-  return NULL;
+{ return NULL;
 }
 
 
@@ -1056,14 +798,12 @@ nbio_last_error(nbio_sock_t socket)
 		 *	  INITIALISATION	*
 		 *******************************/
 
-NBIO_EXPORT(int)
+int
 nbio_init(const char *module)
 { if ( initialised )			/* called from install handlers, which */
     return TRUE;			/* are serialized by the compiler mutex */
   initialised = TRUE;
 
-  INITLOCK();
-
   FUNCTOR_module2  = PL_new_functor(PL_new_atom(":"), 2);
   FUNCTOR_ip4	   = PL_new_functor(PL_new_atom("ip"), 4);
   FUNCTOR_ip1	   = PL_new_functor(PL_new_atom("ip"), 1);
@@ -1083,7 +823,7 @@ nbio_init(const char *module)
 }
 
 
-NBIO_EXPORT(int)
+int
 nbio_cleanup(void)
 { if ( initialised )
   {
@@ -1102,7 +842,7 @@ socket(-Socket)
     the format $socket(Id).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-NBIO_EXPORT(nbio_sock_t)
+nbio_sock_t
 nbio_socket(int domain, int type, int protocol)
 { SOCKET sock;
   plsocket *s;
@@ -1111,65 +851,72 @@ nbio_socket(int domain, int type, int protocol)
 
   if ( (sock = socket(domain, type , protocol)) == INVALID_SOCKET )
   { nbio_error(GET_ERRNO, TCP_ERRNO);
-    return -1;
+    return NULL;
   }
   if ( !(s=allocSocket(sock)) )		/* register it */
   { closesocket(sock);
-    return -1;
+    return NULL;
   }
-
-  return s->id;
+#ifdef __WINDOWS__
+  /* On older versions of Windows (win7 and before) the default send
+     buffer size is only 8k. On a high latency link this can seriously
+     hamper performance. Default it to max(64k, current-value)
+  */
+  { int val;
+    int val_len = sizeof(val);
+    if (getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char*)&val, &val_len) == 0 &&
+	val < 65535)
+    { val = 65535;
+      setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char*)&val, sizeof(int));
+    }
+  }
+#endif
+  return s;
 }
 
 
-NBIO_EXPORT(int)
+int
 nbio_closesocket(nbio_sock_t socket)
-{ plsocket *s;
-  int rc = 0;
+{ int rc = 0;
 
-  if ( !(s = nbio_to_plsocket_raw(socket)) )
-  { DEBUG(1, Sdprintf("nbio_closesocket(%d): no plsocket\n", socket));
-    return -1;
-  }
+  VALID_SOCKET(socket);
 
-  s->flags &= ~PLSOCK_VIRGIN;
+  clear(socket, PLSOCK_VIRGIN);
 
-  if ( true(s, PLSOCK_OUTSTREAM|PLSOCK_INSTREAM) )
-  { int flags = s->flags;		/* may drop out! */
+  if ( true(socket, PLSOCK_OUTSTREAM|PLSOCK_INSTREAM) )
+  { int flags = socket->flags;		/* may drop out! */
 
     if ( flags & PLSOCK_INSTREAM )
-    { assert(s->input);
-      if ( Slock(s->input) == 0 )
-	rc += Sclose(s->input);
+    { assert(socket->input);
+      if ( Slock(socket->input) == 0 )
+	rc += Sclose(socket->input);
       else
 	rc--;
     }
     if ( flags & PLSOCK_OUTSTREAM )
-    { assert(s->output);
-      if ( Slock(s->output) == 0 )
-	rc += Sclose(s->output);
+    { assert(socket->output);
+      if ( Slock(socket->output) == 0 )
+	rc += Sclose(socket->output);
       else
 	rc--;
     }
   } else
   { rc = 0;
 #ifdef __WINDOWS__
-    shutdown(s->socket, SD_BOTH);
+    shutdown(socket->socket, SD_BOTH);
 #endif
-    freeSocket(s);
+    closeSocket(socket);
   }
 
   return rc;
 }
 
-NBIO_EXPORT(int)
+int
 nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
-{ plsocket *s;
-  va_list args;
+{ va_list args;
   int rc;
 
-  if ( !(s = nbio_to_plsocket(socket)) )
-    return -1;
+  VALID_SOCKET(socket);
 
   va_start(args, opt);
 
@@ -1180,9 +927,9 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
     case TCP_REUSEADDR:
     { int val = va_arg(args, int);
 
-      if( setsockopt(s->socket, SOL_SOCKET, SO_REUSEADDR,
+      if( setsockopt(socket->socket, SOL_SOCKET, SO_REUSEADDR,
 		     (const char *)&val, sizeof(val)) == -1 )
-      { nbio_error(GET_H_ERRNO, TCP_HERRNO);
+      { nbio_error(GET_ERRNO, TCP_ERRNO);
 	rc = -1;
       } else
 	rc = 0;
@@ -1193,7 +940,7 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
     { const char *dev = va_arg(args, char*);
 
 #ifdef SO_BINDTODEVICE
-      if ( setsockopt(s->socket, SOL_SOCKET, SO_BINDTODEVICE,
+      if ( setsockopt(socket->socket, SOL_SOCKET, SO_BINDTODEVICE,
 		      dev, strlen(dev)) == 0 )
       { rc = 0;
         break;
@@ -1214,9 +961,9 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 #ifndef IPPROTO_TCP			/* Is this correct? */
 #define IPPROTO_TCP SOL_SOCKET
 #endif
-      if ( setsockopt(s->socket, IPPROTO_TCP, TCP_NODELAY,
+      if ( setsockopt(socket->socket, IPPROTO_TCP, TCP_NODELAY,
 		      (const char *)&val, sizeof(val)) == -1 )
-      { nbio_error(GET_H_ERRNO, TCP_HERRNO);
+      { nbio_error(GET_ERRNO, TCP_ERRNO);
 	rc = -1;
       } else
       { rc = 0;
@@ -1232,9 +979,9 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
     case UDP_BROADCAST:
     { int val = va_arg(args, int);
 
-      if ( setsockopt(s->socket, SOL_SOCKET, SO_BROADCAST,
+      if ( setsockopt(socket->socket, SOL_SOCKET, SO_BROADCAST,
 		     (const char *)&val, sizeof(val)) == -1 )
-      { nbio_error(GET_H_ERRNO, TCP_HERRNO);
+      { nbio_error(GET_ERRNO, TCP_ERRNO);
 	rc = -1;
       } else
 	rc = 0;
@@ -1245,9 +992,9 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
     { int val = va_arg(args, int);
 
       if ( val )
-	set(s, PLSOCK_DISPATCH);
+	set(socket, PLSOCK_DISPATCH);
       else
-	clear(s, PLSOCK_DISPATCH);
+	clear(socket, PLSOCK_DISPATCH);
 
       rc = 0;
 
@@ -1256,9 +1003,11 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
     case TCP_INSTREAM:
     { IOSTREAM *in = va_arg(args, IOSTREAM*);
 
-      s->flags |= PLSOCK_INSTREAM;
-      s->flags &= ~PLSOCK_VIRGIN;
-      s->input = in;
+      socket->flags |= PLSOCK_INSTREAM;
+      socket->flags &= ~PLSOCK_VIRGIN;
+      socket->input = in;
+      if ( socket->symbol )
+	PL_register_atom(socket->symbol);
 
       rc = 0;
 
@@ -1267,14 +1016,28 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
     case TCP_OUTSTREAM:
     { IOSTREAM *out = va_arg(args, IOSTREAM*);
 
-      s->flags |= PLSOCK_OUTSTREAM;
-      s->flags &= ~PLSOCK_VIRGIN;
-      s->output = out;
+      socket->flags |= PLSOCK_OUTSTREAM;
+      socket->flags &= ~PLSOCK_VIRGIN;
+      socket->output = out;
+      if ( socket->symbol )
+	PL_register_atom(socket->symbol);
 
       rc = 0;
 
       break;
     }
+    case TCP_SNDBUF:
+    { int val = va_arg(args, int);
+      if ( setsockopt(socket->socket, SOL_SOCKET, SO_SNDBUF,
+		      (const char *)&val, sizeof(int)) == -1 )
+      { nbio_error(GET_ERRNO, TCP_ERRNO);
+	rc = -1;
+      } else
+      { rc = 0;
+      }
+
+      break;
+    }
     default:
       rc = -1;
       assert(0);
@@ -1288,12 +1051,9 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 
 int
 nbio_get_flags(nbio_sock_t socket)
-{ plsocket *s;
+{ VALID_SOCKET(socket);
 
-  if ( !(s = nbio_to_plsocket(socket)) )
-    return -1;
-
-  return s->flags;
+  return socket->flags;
 }
 
 
@@ -1345,11 +1105,12 @@ nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr, term_t *varport)
     if ( PL_get_atom_chars(arg, &hostName) )
     { struct addrinfo hints;
       struct addrinfo *res;
+      int rc;
 
       memset(&hints, 0, sizeof(hints));
       hints.ai_family = AF_INET;
-      if ( getaddrinfo(hostName, NULL, &hints, &res) !=  0) /* see (*) */
-	return nbio_error(GET_H_ERRNO, TCP_HERRNO);
+      if ( (rc=getaddrinfo(hostName, NULL, &hints, &res)) !=  0) /* see (*) */
+	return nbio_error(rc, TCP_GAI_ERRNO);
       assert(res->ai_family == AF_INET);
       memcpy(&addr->sin_addr,
 	     &((struct sockaddr_in*)res->ai_addr)->sin_addr,
@@ -1430,22 +1191,19 @@ nbio_unify_ip4(term_t Ip, unsigned long hip)
 
 int
 nbio_bind(nbio_sock_t socket, struct sockaddr *my_addr, size_t addrlen)
-{ plsocket *s;
-
-  if ( !(s = nbio_to_plsocket(socket)) )
-    return -1;
+{ VALID_SOCKET(socket);
 
 #ifdef __WINDOWS__
-  if ( bind(s->socket, my_addr, (int)addrlen) )
+  if ( bind(socket->socket, my_addr, (int)addrlen) )
 #else
-  if ( bind(s->socket, my_addr, addrlen) )
+  if ( bind(socket->socket, my_addr, addrlen) )
 #endif
   {
     nbio_error(GET_ERRNO, TCP_ERRNO);
     return -1;
   }
 
-  s->flags |= PLSOCK_BIND;
+  set(socket, PLSOCK_BIND);
 
   return 0;
 }
@@ -1455,18 +1213,15 @@ int
 nbio_connect(nbio_sock_t socket,
 	     const struct sockaddr *serv_addr,
 	     size_t addrlen)
-{ plsocket *s;
-
-  if ( !(s = nbio_to_plsocket(socket)) )
-    return -1;
+{ VALID_SOCKET(socket);
 
   for(;;)
-  { if ( connect(s->socket, serv_addr, addrlen) )
+  { if ( connect(socket->socket, serv_addr, addrlen) )
     { if ( need_retry(GET_ERRNO) )
       { if ( PL_handle_signals() < 0 )
 	  return -1;
 #ifdef __WINDOWS__
-        if ( !wait_socket(s) )
+        if ( !wait_socket(socket) )
           return -1;
 #endif
 	continue;
@@ -1481,7 +1236,7 @@ nbio_connect(nbio_sock_t socket,
       break;
   }
 
-  s->flags |= PLSOCK_CONNECT;
+  set(socket, PLSOCK_CONNECT);
 
   return 0;
 }
@@ -1490,32 +1245,31 @@ nbio_connect(nbio_sock_t socket,
 nbio_sock_t
 nbio_accept(nbio_sock_t master, struct sockaddr *addr, socklen_t *addrlen)
 { SOCKET slave;
-  plsocket *m, *s;
+  plsocket *s;
 
-  if ( !(m = nbio_to_plsocket(master)) )
-    return -1;
+  VALID_SOCKET_RET(master, NULL);
 
   for(;;)
   {
 #ifndef __WINDOWS__
-    if ( !wait_socket(m) )
-      return -1;
+    if ( !wait_socket(master) )
+      return NULL;
 #endif
 
-    slave = accept(m->socket, addr, addrlen);
+    slave = accept(master->socket, addr, addrlen);
 
     if ( slave == SOCKET_ERROR )
     { if ( need_retry(GET_ERRNO) )
       { if ( PL_handle_signals() < 0 )
-	  return -1;
+	  return NULL;
 #ifdef __WINDOWS__
-        if ( !wait_socket(m) )
-          return -1;
+        if ( !wait_socket(master) )
+          return NULL;
 #endif
 	continue;
       } else
       { nbio_error(GET_ERRNO, TCP_ERRNO);
-	return -1;
+	return NULL;
       }
     } else
       break;
@@ -1525,26 +1279,23 @@ nbio_accept(nbio_sock_t master, struct sockaddr *addr, socklen_t *addrlen)
   s->flags |= PLSOCK_ACCEPT;
 #ifndef __WINDOWS__
   if ( true(s, PLSOCK_NONBLOCK) )
-    nbio_setopt(slave, TCP_NONBLOCK);
+    nbio_setopt(s, TCP_NONBLOCK);
 #endif
 
-  return s->id;
+  return s;
 }
 
 
 int
 nbio_listen(nbio_sock_t socket, int backlog)
-{ plsocket *s;
+{ VALID_SOCKET(socket);
 
-  if ( !(s = nbio_to_plsocket(socket)) )
-    return -1;
-
-  if( listen(s->socket, backlog) == -1 )
+  if( listen(socket->socket, backlog) == -1 )
   { nbio_error(GET_ERRNO, TCP_ERRNO);
     return -1;
   }
 
-  s->flags |= PLSOCK_LISTEN;
+  set(socket, PLSOCK_LISTEN);
 
   return 0;
 }
@@ -1557,23 +1308,19 @@ nbio_listen(nbio_sock_t socket, int backlog)
 #define fdFromHandle(p) ((int)((intptr_t)(p)))
 
 ssize_t
-nbio_read(int socket, char *buf, size_t bufSize)
-{ plsocket *s;
-  int n;
+nbio_read(nbio_sock_t socket, char *buf, size_t bufSize)
+{ int n;
 
-  if ( !(s = nbio_to_plsocket(socket)) )
-    return -1;
+  VALID_SOCKET(socket);
 
   for(;;)
   {
 #ifndef __WINDOWS__
-    if ( !wait_socket(s) )
-    { errno = EPLEXCEPTION;
+    if ( !wait_socket(socket) )
       return -1;
-    }
 #endif
 
-    n = recv(s->socket, buf, bufSize, 0);
+    n = recv(socket->socket, buf, bufSize, 0);
 
     if ( n == -1 )
     { if ( need_retry(GET_ERRNO) )
@@ -1582,7 +1329,7 @@ nbio_read(int socket, char *buf, size_t bufSize)
           return -1;
         }
 #ifdef __WINDOWS__
-        if ( !wait_socket(s) )
+        if ( !wait_socket(socket) )
           return -1;
 #endif
         continue;
@@ -1600,17 +1347,15 @@ nbio_read(int socket, char *buf, size_t bufSize)
 
 ssize_t
 nbio_write(nbio_sock_t socket, char *buf, size_t bufSize)
-{ plsocket *s;
-  size_t len = bufSize;
+{ size_t len = bufSize;
   char *str = buf;
 
-  if ( !(s = nbio_to_plsocket(socket)) )
-    return -1;
+  VALID_SOCKET(socket);
 
   while( len > 0 )
   { int n;
 
-    n = send(s->socket, str, len, 0);
+    n = send(socket->socket, str, len, 0);
     if ( n < 0 )
     { if ( need_retry(GET_ERRNO) )
       { if ( PL_handle_signals() < 0 )
@@ -1618,7 +1363,7 @@ nbio_write(nbio_sock_t socket, char *buf, size_t bufSize)
 	  return -1;
 	}
 #ifdef __WINDOWS__
-        if ( !wait_socket(s) )
+        if ( !wait_socket(socket) )
           return -1;
 #endif
 	continue;
@@ -1641,50 +1386,85 @@ nbio_write(nbio_sock_t socket, char *buf, size_t bufSize)
 }
 
 
+#if defined(__WINDOWS__) && !defined(SHUT_RD)
+#define SHUT_RD SD_RECEIVE
+#define SHUT_WR SD_SEND
+#endif
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+According to [1], we need to run:
+
+ 1. Call WSAEventSelect to register for FD_CLOSE notification.
+ 2. Call shutdown with how=SD_SEND.
+ 3. When FD_CLOSE received, call the recv or WSARecv until the function
+    completes with success and indicates that zero bytes were received.
+    If SOCKET_ERROR is returned, then the graceful disconnect is not
+    possible.
+ 4. Call closesocket.
+
+We  operationalize  this   by   using    shutdown()   for   sending   in
+nbio_close_output(). A well behaved client  should subsequently read the
+input until EOF. That, in the Windows case, will also consume FD_CLOSE.
+
+[1]
+https://docs.microsoft.com/en-us/windows/desktop/api/winsock/nf-winsock-shutdown
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
 nbio_close_input(nbio_sock_t socket)
 { int rc = 0;
-  plsocket *s;
 
-  if ( !(s = nbio_to_plsocket_raw(socket)) )
-    return -1;
+  VALID_SOCKET(socket);
 
   DEBUG(2, Sdprintf("[%d]: nbio_close_input(%d, flags=0x%x)\n",
-		    PL_thread_self(), socket, s->flags));
-  s->flags &= ~PLSOCK_INSTREAM;
+		    PL_thread_self(), socket, socket->flags));
+  if ( true(socket, PLSOCK_INSTREAM) )
+  { clear(socket, PLSOCK_INSTREAM);
 
-  s->input = NULL;
-  if ( !(s->flags & (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
-  {
-#ifdef __WINDOWS__
-    return shutdown(s->socket, SD_BOTH);
-#endif
-    return freeSocket(s);
+    socket->input = NULL;
+    if ( false(socket, (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
+      rc = closeSocket(socket);
+
+    if ( socket->symbol )
+      PL_unregister_atom(socket->symbol);
   }
 
   return rc;
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+(*) This gives an endpoint not connected   error  for the http:proxy and
+http:websocket tests. I'm not sure whether the   test  is broken or not,
+but error checking will probably cause many  applications to fail and it
+is not that clear that it is an error if the other side closed its input
+and we only want to send TCP FIN.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
 nbio_close_output(nbio_sock_t socket)
 { int rc = 0;
-  plsocket *s;
 
-  if ( !(s = nbio_to_plsocket_raw(socket)) )
-    return -1;
+  VALID_SOCKET(socket);
 
   DEBUG(2, Sdprintf("[%d]: nbio_close_output(%d, flags=0x%x)\n",
-		    PL_thread_self(), socket, s->flags));
-  s->flags &= ~PLSOCK_OUTSTREAM;
+		    PL_thread_self(), socket, socket->flags));
 
-  s->output = NULL;
-  if ( !(s->flags & (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
-  {
-#ifdef __WINDOWS__
-    return shutdown(s->socket, SD_BOTH);
-#endif
-    return freeSocket(s);
+  if ( true(socket, PLSOCK_OUTSTREAM) )
+  { clear(socket, PLSOCK_OUTSTREAM);
+
+    if ( socket->socket != INVALID_SOCKET )
+    { /* if ( (rc = shutdown(socket->socket, SHUT_WR)) )
+	nbio_error(GET_ERRNO, TCP_ERRNO);		See (*) */
+      shutdown(socket->socket, SHUT_WR);
+    }
+
+    socket->output = NULL;
+    if ( false(socket, (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
+      rc = (rc + closeSocket(socket)) ? -1 : 0;
+
+    if ( socket->symbol )
+      PL_unregister_atom(socket->symbol);
   }
 
   return rc;
@@ -1696,24 +1476,20 @@ nbio_close_output(nbio_sock_t socket)
 		 *******************************/
 
 ssize_t
-nbio_recvfrom(int socket, void *buf, size_t bufSize, int flags,
+nbio_recvfrom(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
 	     struct sockaddr *from, socklen_t *fromlen)
-{ plsocket *s;
-  int n;
+{ int n;
 
-  if ( !(s = nbio_to_plsocket(socket)) )
-    return -1;
+  VALID_SOCKET(socket);
 
   for(;;)
   {
 #ifndef __WINDOWS__
-    if ( (flags & MSG_DONTWAIT) == 0 && !wait_socket(s) )
-    { errno = EPLEXCEPTION;
+    if ( (flags & MSG_DONTWAIT) == 0 && !wait_socket(socket) )
       return -1;
-    }
 #endif
 
-    n = recvfrom(s->socket, buf, bufSize, flags, from, fromlen);
+    n = recvfrom(socket->socket, buf, bufSize, flags, from, fromlen);
 
     if ( n == -1 )
     { if ( need_retry(GET_ERRNO) )
@@ -1722,7 +1498,7 @@ nbio_recvfrom(int socket, void *buf, size_t bufSize, int flags,
           return -1;
         }
 #ifdef __WINDOWS__
-        if ( !wait_socket(s) )
+        if ( !wait_socket(socket) )
           return -1;
 #else
         if((flags & MSG_DONTWAIT) != 0)
@@ -1744,14 +1520,12 @@ nbio_recvfrom(int socket, void *buf, size_t bufSize, int flags,
 ssize_t
 nbio_sendto(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
 	    const struct sockaddr *to, socklen_t tolen)
-{ plsocket *s;
-  ssize_t n;
+{ ssize_t n;
 
-  if ( !(s = nbio_to_plsocket(socket)) )
-    return -1;
+  VALID_SOCKET(socket);
 
   for(;;)
-  { n = sendto(s->socket, buf, bufSize, flags, to, tolen);
+  { n = sendto(socket->socket, buf, bufSize, flags, to, tolen);
 
     if ( n < 0 )
     { if ( need_retry(GET_ERRNO) )
@@ -1760,7 +1534,7 @@ nbio_sendto(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
 	  return -1;
 	}
 #ifdef __WINDOWS__
-        if ( !wait_socket(s) )
+        if ( !wait_socket(socket) )
           return -1;
 #endif
         continue;
