@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2007-2019, University of Amsterdam
+    Copyright (c)  2007-2020, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
     All rights reserved.
@@ -34,8 +34,6 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
-:- if(current_predicate(is_dict/1)).
-
 :- module(json,
           [ json_read/2,                % +Stream, -JSONTerm
             json_read/3,                % +Stream, -JSONTerm, +Options
@@ -51,21 +49,6 @@
             json_write_dict/3,          % +Stream, +Dict, +Options
             atom_json_dict/3            % ?Atom, ?JSONDict, +Options
           ]).
-
-:- else.
-
-:- module(json,
-          [ json_read/2,                % +Stream, -JSONTerm
-            json_read/3,                % +Stream, -JSONTerm, +Options
-            atom_json_term/3,           % ?Atom, ?JSONTerm, +Options
-            json_write/2,               % +Stream, +Term
-            json_write/3,               % +Stream, +Term, +Options
-            is_json_term/1,             % @Term
-            is_json_term/2
-          ]).
-
-:- endif.
-
 :- use_module(library(record)).
 :- use_module(library(memfile)).
 :- use_module(library(error)).
@@ -478,19 +461,23 @@ stream_error_context(Stream, stream(Stream, Line, LinePos, CharNo)) :-
 %!  json_write(+Stream, +Term) is det.
 %!  json_write(+Stream, +Term, +Options) is det.
 %
-%   Write a JSON term to Stream.  The   JSON  object  is of the same
-%   format as produced by json_read/2, though we allow for some more
-%   flexibility with regard to pairs in  objects. All of Name=Value,
-%   Name-Value and Name(Value) produce the  same output.
+%   Write a JSON term to Stream. The JSON   object is of the same format
+%   as  produced  by  json_read/2,  though  we    allow  for  some  more
+%   flexibility with regard to pairs  in   objects.  All  of Name=Value,
+%   Name-Value and Name(Value) produce the same output.
 %
-%   Values can be of the form  #(Term),   which  causes `Term` to be
-%   _stringified_ if it is not an atom or string. Stringification is
+%   Values can be of  the  form  #(Term),   which  causes  `Term`  to be
+%   _stringified_ if it is not  an   atom  or string. Stringification is
 %   based on term_string/2.
 %
-%   The version 7 _dict_ type is supported as well.  Optionally,  if
-%   the dict has a _tag_,  a property  "type":"tag"  can be added to
-%   the object.  This behaviour can be controlled using the  =tag=
-%   option (see below). For example:
+%   Rational numbers are emitted as  floating   point  numbers. The hook
+%   json_write_hook/4  can  be  used   to    realize   domain   specific
+%   alternatives.
+%
+%   The version 7 _dict_ type is supported   as well. Optionally, if the
+%   dict has a _tag_, a  property  "type":"tag"   can  be  added  to the
+%   object. This behaviour can be controlled using the =tag= option (see
+%   below). For example:
 %
 %     ==
 %     ?- json_write(current_output, point{x:1,y:2}).
@@ -578,7 +565,6 @@ json_write_term(Var, _, _, _) :-
 json_write_term(json(Pairs), Stream, State, Options) :-
     !,
     json_write_object(Pairs, Stream, State, Options).
-:- if(current_predicate(is_dict/1)).
 json_write_term(Dict, Stream, State, Options) :-
     is_dict(Dict),
     !,
@@ -590,7 +576,6 @@ json_write_term(Dict, Stream, State, Options) :-
     ;   Pairs = Pairs0
     ),
     json_write_object(Pairs, Stream, State, Options).
-:- endif.
 json_write_term(List, Stream, State, Options) :-
     is_list(List),
     !,
@@ -617,7 +602,13 @@ json_write_term(Term, Stream, State, Options) :-
 json_write_term(Number, Stream, _State, _Options) :-
     number(Number),
     !,
-    write(Stream, Number).
+    (   float(Number)
+    ->  write(Stream, Number)
+    ;   integer(Number)
+    ->  write(Stream, Number)
+    ;   Float is float(Number)              % rational number
+    ->  write(Stream, Float)
+    ).
 json_write_term(True, Stream, _State, Options) :-
     json_options_true(Options, True),
     !,
@@ -772,7 +763,6 @@ json_print_length(json(Pairs), Options, Max, Len0, Len) :-
     Len1 =< Max,
     must_be(list, Pairs),
     pairs_print_length(Pairs, Options, Max, Len1, Len).
-:- if(current_predicate(is_dict/1)).
 json_print_length(Dict, Options, Max, Len0, Len) :-
     is_dict(Dict),
     !,
@@ -780,7 +770,6 @@ json_print_length(Dict, Options, Max, Len0, Len) :-
     Len1 is Len0 + 2,
     Len1 =< Max,
     pairs_print_length(Pairs, Options, Max, Len1, Len).
-:- endif.
 json_print_length(Array, Options, Max, Len0, Len) :-
     is_list(Array),
     !,
@@ -916,8 +905,6 @@ is_json_pair(Options, Name=Value) :-
     atom(Name),
     is_json_term2(Options, Value).
 
-:- if(current_predicate(is_dict/1)).
-
                  /*******************************
                  *         DICT SUPPORT         *
                  *******************************/
@@ -1055,9 +1042,7 @@ atom_json_dict(Atom, Term, Options) :-
     ground(Atom),
     !,
     setup_call_cleanup(
-        ( text_memfile(Atom, MF),
-          open_memory_file(MF, read, In, [free_on_close(true)])
-        ),
+        open_string(Atom, In),
         json_read_dict(In, Term, Options),
         close(In)).
 atom_json_dict(Result, Term, Options) :-
@@ -1069,17 +1054,6 @@ atom_json_dict(Result, Term, Options) :-
     with_output_to(Out,
                    json_write_dict(current_output, Term, Options1)).
 
-text_memfile(Atom, MF) :-
-    atom(Atom),
-    !,
-    atom_to_memory_file(Atom, MF).
-text_memfile(String, MF) :-
-    string(String),
-    !,
-    new_memory_file(MF),
-    insert_memory_file(MF, 0, String).
-
-:- endif.
 
                  /*******************************
                  *           MESSAGES           *
