@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2009-2014, University of Amsterdam
+    Copyright (c)  2009-2020, University of Amsterdam
                               VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -40,19 +41,38 @@
             rdf_save_canonical_trig/2,          % +File, +Options
             rdf_save_ntriples/2                 % +File, +Options
           ]).
-:- use_module(library(semweb/rdf_db)).
+:- use_module(library(record),[(record)/1, op(_,_,record)]).
 :- use_module(library(semweb/turtle), []). % we make calls to public preds here
-:- use_module(library(option)).
-:- use_module(library(record)).
-:- use_module(library(error)).
-:- use_module(library(lists)).
-:- use_module(library(rbtrees)).
-:- use_module(library(apply)).
-:- use_module(library(url)).
-:- use_module(library(pairs)).
-:- use_module(library(debug)).
-:- use_module(library(sgml_write)).
-:- use_module(library(sgml)).
+:- use_module(library(semweb/rdf_db),
+              [ rdf_graph/1, rdf_graph_prefixes/3, rdf_current_prefix/2,
+                rdf_is_bnode/1, rdf_equal/2, rdf_graph_property/2,
+                rdf_statistics/1, rdf/4, rdf_resource/1, rdf_subject/1,
+                rdf/3, rdf_global_id/2
+              ]).
+
+:- autoload(library(apply),[maplist/3,include/3,partition/4]).
+:- autoload(library(debug),[assertion/1]).
+:- autoload(library(error),[must_be/2,existence_error/2,type_error/2]).
+:- autoload(library(lists),
+	    [append/2,reverse/2,delete/3,append/3,select/3,member/2]).
+:- autoload(library(option),[meta_options/3]).
+:- autoload(library(pairs),
+	    [ transpose_pairs/2,
+	      map_list_to_pairs/3,
+	      pairs_values/2,
+	      group_pairs_by_key/2
+	    ]).
+:- autoload(library(rbtrees),
+	    [ ord_list_to_rbtree/2,
+	      rb_lookup/3,
+	      rb_insert/4,
+	      rb_empty/1,
+	      rb_update/5
+	    ]).
+:- autoload(library(sgml),
+	    [xml_name/1,xml_is_dom/1,xsd_number_string/2]).
+:- autoload(library(sgml_write),[xml_write/2]).
+:- autoload(library(url),[file_name_to_url/2,parse_url/2]).
 
 :- predicate_options(rdf_save_turtle/2, 2,
                      [ graph(atom),
@@ -151,6 +171,7 @@ has the following properties:
 :- meta_predicate
     rdf_save_turtle(+, :),
     rdf_save_canonical_turtle(+, :),
+    rdf_save_canonical_trig(+, :),
     rdf_save_trig(+, :).
 
 %!  rdf_save_turtle(+Out, :Options) is det.
@@ -362,7 +383,7 @@ tw_trig_graphs([H|T], Stream, State0, State) :-
     set_graph_of_tw_state(H, State0, State1),
     nl(Stream),
     tw_resource(H, State1, Stream),
-    format(Stream, ' = {~n', []),
+    format(Stream, ' {~n', []),
     tw_graph(State1, Stream),
     format(Stream, '~N}~n', []),
     set_bnode_id_of_tw_state(0, State1, State2),
@@ -469,7 +490,8 @@ graph_prefixes(State0, Graph, Prefixes) :-
     rdf_graph_prefixes(Graph, Prefixes,
                        [ filter(turtle_prefix(OnlyKnown)),
                          expand(Expand),
-                         min_count(2)
+                         min_count(2),
+                         get_prefix(turtle:iri_turtle_prefix)
                        ]).
 
 prefix_map(State, Prefixes, PrefixMap) :-
@@ -596,7 +618,7 @@ propose_abbrev(_, Len, URI, Abbrev) :-
     atomic_list_concat(Use, -, Abbrev).
 
 abbrev_part(X) :-
-    turtle:turtle_pn_local(X),
+    xml_name(X),
     \+ well_known_ns(X, _),
     \+ well_known_extension(X).
 
@@ -1341,7 +1363,8 @@ tw_resource(Resource, State, Out) :-
     ;   Name == ''
     ),
     !,
-    format(Out, '~w:~w', [Prefix, Name]).
+    format(Out, '~w:', [Prefix]),
+    turtle:turtle_write_pn_local(Out, Name).
 tw_resource(Resource, State, Out) :-
     tw_relative_uri(Resource, State, Out).
 
@@ -1513,11 +1536,25 @@ inc_subject_count(State, Count) :-
     prolog:message//1.
 
 prolog:message(rdf(saved(File, Time, SavedSubjects, SavedTriples))) -->
-    [ 'Saved ~D triples about ~D subjects into ~p (~3f sec)'-
-      [SavedTriples, SavedSubjects, File, Time]
-    ].
+    [ 'Saved ~D triples about ~D subjects into '-[SavedTriples, SavedSubjects] ],
+    rdf_output(File),
+    [ ' (~3f sec)'-[Time] ].
 prolog:message(rdf(saved(File, Time, SavedSubjects, SavedTriples,
                          SavedGraphs))) -->
-    [ 'Saved ~D graphs, ~D triples about ~D subjects into ~p (~3f sec)'-
-      [SavedGraphs, SavedTriples, SavedSubjects, File, Time]
-    ].
+    [ 'Saved ~D graphs, ~D triples about ~D subjects into '-
+      [SavedGraphs, SavedTriples, SavedSubjects] ],
+    rdf_output(File),
+    [ ' (~3f sec)'-[Time] ].
+
+rdf_output(StreamSpec) -->
+    { (   StreamSpec = stream(Stream)
+      ->  true
+      ;   Stream = StreamSpec
+      ),
+      is_stream(Stream),
+      stream_property(Stream, file_name(File))
+    },
+    !,
+    [ '~p'-[File] ].
+rdf_output(File) -->
+    [ '~p'-[File] ].
