@@ -179,6 +179,8 @@ const debug_topic debug_topics[] =
 
   DEBUG_TOPIC(MSG_ZIP),
   DEBUG_TOPIC(MSG_ZIP_STATE),
+  DEBUG_TOPIC(MSG_WIN_API),
+  DEBUG_TOPIC(MSG_WIN_DDE),
 
   DEBUG_TOPIC(MSG_MODULE_REF),
 
